#!/usr/bin/perl
#
# rpmcheck.pl
#
# Copyright JulianDyke.com Limited 2010-2015
# 
# This package compares a package specification file supplied in XML format 
# to the packages currently installed in Linux
#
# Version History
# 09Nov10 JMD Version 1.0
#   Replaced XML::Simple with custom XML parser
# 15Nov10 JMD Version 1.1
#   Included strict module - updated parameters
#   Fixed logic errors in compareStrings()
# 26Nov10 JMD Version 1.2
#   Changed Oracle versions in XML file names to major release only
#   e.g. from 11.2.0.2 to 11.2
# 02Dec10 JMD Version 1.3
#   Changed Oracle versions within XML files to major release only
#   e.g. from 11.2.0.2 to 11.2
#   Removed numactl-devel package from all XML files as it does not appear
#   to be required
# 23Feb10 JMD Version 1.5
#   Introduced compareReleases procedure to simply code
#   Changed check_package to call compareReleases instead of compareVersion
#   for to compare release numbers as algorithm is different
#   Attempts to fix bug where
#     glibc/2.5/24 should be older than glibc/2.5/49.el5_5.7
# 16Jun11 JMD Version 1.6
#   Fixed version numbers in GridInfrastructure 11.2 XML files for 
#   libaio-devel package
# 04Aug14 JMD Version 1.7
#   Added Database-11.2-OEL6-x64_64.xml
#   Added Database-12.1-OEL6-x64_64.xml
#   Added Database-11.2-RHEL6-x64_64.xml
#   Added Database-12.1-RHEL6-x64_64.xml

use strict;

my $trace = 0;
my $numPackages = 0;
my $numFailures = 0;    # Assume a success until we fail
my %product = {};
my $filename = "";
my $numArgs = 0;
my @package;
my $ch = "";
my $eof = 0;
my $xml = "";
my $iptr = 0;

# get next character
sub getch 
{
  $ch = substr ($xml,$iptr,1);
  $iptr++;
  if ($iptr == length ($xml))
  {
    $eof = 1;
  }
}

# parse XML document
sub parseDocument 
{
  my $level = 0;    # 1 = product 2 = package
  my $key = " ";

  while (<F>)
  {
    # remove newlines
    chomp;
    my $line = $_;
    # remove carriage returns
    $line =~ tr/\r//d;
    $xml = $xml . $line;
  }
  
  # get first character
  getch ();
  
  # while not end of input
  while ($eof == 0)
  {
    # skip spaces
    while ($ch eq " ")
    {
      getch ();
    }
  
    # if first character is "<"
    if ($ch eq "<")
    {
      # get next character
      getch ();
  
      # if character is "?" 
      if ($ch eq "?")
      {
        # it is a directive - ignore
        while ($ch ne ">")
        {
          getch ();
        }
        getch ();
      }
      else   # Not a directive 
      {
        # If next character is a / then it is an end tag
        if ($ch eq "/")
        {
          # end tag
          getch ();
          my $tag = "";
          until ($ch eq ">")
          {
            $tag = $tag . $ch;
            getch ();
          }
 
          if ($trace == 1)
          { 
            printf ("End Tag = %s\n",$tag);
          }
          if ($tag eq "package")
          {
            $numPackages++;
          }
        }
        else # Must be a start tag
        {
          my $tag = "";
          until ($ch eq ">")
          {
            $tag = $tag . $ch;
            getch ();
          }
          if ($trace == 1) 
          {
            printf ("Start Tag = %s\n",$tag);
          }
          if ($tag eq "product") 
          {
            $level = 1;
          }
          elsif ($tag eq "package")
          {
            $level = 2;
          }
          else 
          { 
            $key = $tag;
          }
        }
        #  skip over close bracket
        getch ();
      }
    }
    else
    { 
      # data
      my $data = "";
      while ($ch ne "<")
      {
        $data = $data . $ch;
        getch ();
      }      
      if ($trace == 1)
      {
        printf ("Data = %s\n",$data);
      }
      if ($level == 1)
      {
        $product{$key} = $data;
      }
      else
      {
        $package[$numPackages]{$key} = $data;
      }
    }
  }    
}

# subroutine to check if variable is a number
# there are lots of ways to do this in Perl; this is one of the simplest
sub isNumber
{
  if ($_[0] =~ /\D/) {return 0;} else { return 1};
}

sub printHeader 
{
  printf 
  ("Package Name                   Version          Release          Architecture\n");
  printf 
  ("------------------------------ ---------------- ---------------- ------------\n");
  # 123456789012345678901234567890 1234567890123456 1234567890123456 123456789012
}

# compare two strings - may contain numbers and/or characters
sub compareStrings
{
  # sort out the parameters
  my ($a1,$a2) = @_;

  # convert them into arrays
  my @a1 = @$a1;    
  my @a2 = @$a2;

  # get length of shortest (first) string
  my $len = @a1;
 
  # for each element in the shortest string
  for (my $i = 0;$i < $len;$i++)
  {
    # if both strings are numbers
    if ((isNumber ($a1[$i])) && (isNumber ($a2[$i])))
    {
      # if first string is newer than second string return 1
      if ($a1[$i] > $a2[$i]) { return 1 };
    }
    else # at least one string is not a number
    {
      # if first string is newer than second string return 1
      if ($a1[$i] gt $a2[$i]) { return 1 };
    }
  }
 
  # first string is older than (or equal to) second string  
  return -1;
}

# compare two strings containing version numbers
sub compareVersions
{
  my $str1 = $_[0];
  my $str2 = $_[1];

  # Check for equality
  if ($str1 eq $str2)
  {
    # Strings are identical;
    return 0;
  }

  # Convert hyphens to periods
  $str1 =~ tr/\\-/\\./;
  $str2 =~ tr/\\-/\\./;

  # Split string into array at periods
  my @s1 = split (/\./,$str1);
  my @s2 = split (/\./,$str2);

  # Get the array lengths  
  my $len1 = @s1;
  my $len2 = @s2;

  # If first array is shorter than second array
  if ($len1 < $len2)
  {
    # Return result of comparing the arrays
    return (compareStrings (\@s1,\@s2));
  }
  else # First array is same length or longer than second array
  {
    # Return negated result of comparing the arrays
    return (compareStrings (\@s2,\@s1));
  }
}
    
# compare two strings containing release numbers
sub compareReleases
{
  my $str1 = $_[0];
  my $str2 = $_[1];

  # Check for equality
  if ($str1 eq $str2)
  {
    # Strings are identical;
    return 0;
  }

  # Convert hyphens to periods
  $str1 =~ tr/\\-/\\./;
  $str2 =~ tr/\\-/\\./;

  # Split string into array at periods
  my @s1 = split (/\./,$str1);
  my @s2 = split (/\./,$str2);

  # Get the array lengths  
  my $len1 = @s1;
  my $len2 = @s2;

  # If first array is shorter than second array
  if ($len1 < $len2)
  {
    # Return result of comparing the arrays
    return (compareStrings (\@s1,\@s2));
  }
  else # First array is same length or longer than second array
  {
    # Return negated result of comparing the arrays
    return (compareStrings (\@s1,\@s2));
  }
}
    
sub check_package
{
  my $name = $_[0];
  my $version = $_[1];
  my $release = $_[2];
  my $arch = $_[3];

  if ($trace == 1)
  {
    printf ("Checking for package: %s %s %s %s\n",
    $name,$version,$release,$arch);
  }

  # Build the rpm command
  my $rpm = "rpm -q --qf '%{NAME};%{VERSION};%{RELEASE};%{ARCH};\n' " . $name . " |";

  # Execute the rpm command and pipe the output to stdout
  open (RPM,$rpm) || die "Failed: $!\n";
 
  my $found = 0;

  # rpm command may return multiple lines (for different architectures) 
  while (<RPM>)
  {
    # If package not installed 
    if (/is not installed/)
    {
      if ($trace == 1)
      {
        printf ("  Package %s not installed\n",$name);
      }
      # jump out of loop
      last;
    }

    my $n = "";
    my $v = "";
    my $r = "";
    my $a = "";

    # split the rpm output into constituent parts
    ($n,$v,$r,$a) = split (/;/);
  
    if ($trace == 1)
    {
      printf ("  Found package %s %s %s %s\n",$n,$v,$r,$a);
    }
  
    # Sanity check - is this the correct name?
    if ($n eq $name)
    {
      # Is this the target architecture?
      if ($a eq $arch)
      {
        $found = 1;

        # Compare the version numbers
        my $res = compareVersions ($version,$v);        
        if ($res == -1)
        {
          $numFailures++;
          printf ("Package %s is not valid\n",$name);
          printf ("  Expected version %s - Actual version %s\n",$version,$v);
          printf ("  Install %s %s %s (%s) or above\n", 
            $name,$version,$release,$arch);
          last;
        }

        if ($res == 0) # versions are identical
        {
          # if $release field was not empty
          if ($release)
          {
            # Compare the releases 
            $res = compareReleases ($release,$r);        
            if ($res == 1)
            {
              $numFailures++;

              printf ("Package %s is not valid\n",$name); 
              printf ("  Expected release %s - Actual release %s\n",$release,$r);
              printf ("  Install %s %s %s (%s) or above\n", 
                $name,$version,$release,$arch);
              last;
            }
          }
        } 
        # Package must be valid
        printf ("%-30s %-16s %-16s %12s\n",$name,$v,$r,$arch);
        if ($trace == 1)
        {
          printf ("%s;%s;%s;%s;%s;%s;\n",$name,$arch,$version,$release,$v,$r);
        }
        last;
      }
    }
  }

  if ($found == 0)
  { 
    $numFailures++;
    printf ("Package %s not found for architecture %s\n",$name,$arch);
    printf ("  Expected version %s - Expected release %s\n",$version,$release);
    printf ("  Install %s %s %s (%s) or above\n",
      $name,$version,$release,$arch);
  }

  if ($trace == 1)
  {
    printf ("\n");
  }

  close (RPM);

}

# need some argument processing
$numArgs=@ARGV;

if ($numArgs < 1)
{
  printf ("Syntax is rpmcheck.pl <XMLfilename>\n");
  exit 1;
}

$filename=@ARGV[0];
if (!open (F,$filename))
{
  printf ("Failed to open XML file $filename");
  exit 1;
}

printf ("rpmcheck version 1.7\n");
printf ("====================\n"); 
printf ("Copyright JulianDyke.com Limited 2010-2015\n");
printf ("\n");

printf ("Processing %s\n",$filename);

printf ("\n");


parseDocument ();

printf ("Product       : %s\n",$product {"name"});
printf ("Version       : %s\n",$product {"version"});
printf ("Platform      : %s\n",$product {"platform"});
printf ("Arch          : %s\n",$product {"arch"});

printf ("\n");

printf ("Expected packages:\n");

printf ("\n");

printHeader ();

# for each package in the XML document
for (my $i = 0;$i < $numPackages;$i++)
{
  # print the expected version
  printf ("%-30s %-16s %-16s %12s\n",
    $package[$i]->{"name"},
    $package[$i]->{"version"},
    $package[$i]->{"release"},
    $package[$i]->{"arch"}
  );
}
 
printf ("\n");

printf ("Actual packages:\n");

printf ("\n");

# for each package in the XML document
for (my $i = 0;$i < $numPackages;$i++)
{
  # check the actual version
  check_package 
  (
    $package[$i]->{"name"},
    $package[$i]->{"version"},
    $package[$i]->{"release"},
    $package[$i]->{"arch"}
  );
}
 
printf ("\n");

if ($numFailures == 0)
{
  printf ("Package check successful\n");
}
else
{
  if ($numFailures == 1)
  {
    printf ("%d invalid package found\n",$numFailures);
  }
  else
  {
    printf ("%d invalid packages found\n",$numFailures);
  }
}

printf ("\n");

